
/*
 * Program: Watchdog program for the Berkshire Products Internal-Serial
 * Watchdog in Command mode
 * 
 * Copyright (C) 1999/2000  A. Schiffler/Deskplayer Inc., GPL
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free 
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * To contact the author try: Andreas Schiffler
 * aschiffler@home.com Senior Systems Engineer  -  Deskplayer.Com Inc.,
 * Buffalo 4707 Eastwood Cres., Niagara Falls, Ont  L2E 1B4, Canada
 * +1-905-371-3652 (private)  -  +1-905-371-8834 (work/fax)
 * 
 */

/*
 * =========== Includes and watchdog defines 
 */

#include "dpwatch.h"

/*
 * =========== Defines 
 */

/*
 * Filename lengths 
 */

#define MAX_FILENAME_LEN	256

/*
 * Default time(s)/counts 
 */

#define LOOP_DELAY_DEFAULT			5
#define LOOPS_BEFORE_STATUS_DEFAULT		120
#define ERROR_LIMIT_DEFAULT			128
#define READ_ERROR_LIMIT			128

/*
 * =========== Globals 
 */

/*
 * Delay between commands (in seconds) 
 */


static unsigned int LOOP_DELAY = LOOP_DELAY_DEFAULT;


/*
 * How many loops before we get a status printout again 
 */


static int LOOPS_BEFORE_STATUS = LOOPS_BEFORE_STATUS_DEFAULT;


/*
 * How many errors before we exit the loop 
 */


static int ERROR_LIMIT = ERROR_LIMIT_DEFAULT;


/*
 * Reset machine right now (special mode) 
 */

static int RESET_NOW = 0;



/*
 * Other globals 
 */

static struct termios oldtio;	/* Backup of termial setting */
static int in_loop;		/* Main loop control */

/*
 * ============= Procedures 
 */


/* Signal handler for quitting */
void
Quit (int dummy)
{
  /* Quit loop */
  in_loop = 0;
}


/*
 * Create command buffer 
 */

void
CreateSendbuffer (unsigned char command, unsigned char *sendbuffer)
{
  sendbuffer[0] = 0x01;
  sendbuffer[1] = 0x57;
  sendbuffer[2] = 0x84;
  sendbuffer[3] = command;
  sendbuffer[4] =
    (unsigned char) 0x00 - (unsigned char) (0x01 + 0x57 + 0x84 +
					    sendbuffer[3]);
#ifdef DEBUG
  printf ("sendbuffer[] prepared with command 0x%x (%i)\n", command, command);
#endif
}

/*
 * Serial message loop 
 */

void
ResetLoop (int fd)
{
  static unsigned char sendbuffer[WATCHDOG_SENDBUFFER_SIZE];
  static unsigned char recbuffer[WATCHDOG_RECBUFFER_SIZE];
  int loopcounter;
  int errorcounter;
  int readerrors;
  int result;
  unsigned char recvalue = 0;

#ifdef DEBUG
  printf ("ResetLoop() entered, fd=%i\n", fd);
#endif

  /*
   * Reset loop variables 
   */
  in_loop = 1;
  /* Send a few NOPs first */
  loopcounter = -3;	
  errorcounter = 0;
  readerrors = 0;
  while (in_loop)
    {

      /*
       * Write command 
       */
      if ((loopcounter>=0) && (loopcounter < 5)) 
	{
	  /*
	   * Send special commands 
	   */
	  switch (loopcounter)
	    {
	    case 0:
	      /*
	       * Commands that run only during the first invocation 
	       */
	      if (RESET_NOW)
		{
		  syslog (LOG_USER | LOG_WARNING, "Resetting in 10 sec!");
		  CreateSendbuffer (WATCHDOG_RESET_SOON, sendbuffer);
		}
	      else
		{
		  CreateSendbuffer (WATCHDOG_ARM, sendbuffer);
		}
	      break;
	    case 1:
	      CreateSendbuffer (WATCHDOG_REV_MAJOR, sendbuffer);
	      break;
	    case 2:
	      CreateSendbuffer (WATCHDOG_REV_MINOR, sendbuffer);
	      break;
	    case 3:
	      CreateSendbuffer (WATCHDOG_RESETCOUNT, sendbuffer);
	      break;
	    case 4:
	      CreateSendbuffer (WATCHDOG_TEMPERATURE, sendbuffer);
	      break;
	    }
	}
      else
	{
	  /*
	   * Send NoOp 
	   */
	  CreateSendbuffer (WATCHDOG_NOP, sendbuffer);
	}

      /*
       * Write command 
       */
      result = write (fd, sendbuffer, WATCHDOG_SENDBUFFER_SIZE);
#ifdef DEBUG
      printf ("write() returned %i\n", result);
#endif

      /*
       * Pause for a bit 
       */
      usleep (100000);

      /*
       * Read response 
       */
      memset (recbuffer, 0, 4);
      result = read (fd, recbuffer, 4);
#ifdef DEBUG
      printf ("read() returned %i byte(s).\n", result);
#endif

      if (result != 4)
	{
	  readerrors++;
#ifdef DEBUG
	  printf ("Currently %i read errors.\n", readerrors);
#endif
	}

      /*
       * Check read errors 
       */
      if (readerrors == READ_ERROR_LIMIT)
	{
#ifdef DEBUG
	  printf ("Too many read errors.\n");
#endif
	  /*
	   * Too many read errors - device probably not connected. Exit. 
	   */
	  syslog (LOG_USER | LOG_ERR, "Can't read from device ");

	}
      else
	{

	  /*
	   * Analyse the result 
	   */

	  /*
	   * Analyse the response 
	   */
	  /*
	   * Check magic bytes 
	   */
	  if ((recbuffer[0] != 0x57) || (recbuffer[1] != 0x84)
	      || (recbuffer[3] != 0x0d))
	    {
#ifdef DEBUG
	      printf ("Magic bytes corrupted.\n");
#endif
	      syslog (LOG_USER | LOG_WARNING,
		      "Problem - result magic bytes are corrupted.");
	      errorcounter++;
	    }

	  /*
	   * Check result code 
	   */
	  if (recbuffer[2] == 0x06)
	    {
	      /*
	       * Everything OK 
	       */
#ifdef DEBUG
	      printf ("ACK received\n");
#endif
	    }
	  else if (recbuffer[2] == 0x15)
	    {
#ifdef DEBUG
	      printf ("NAK1 received\n");
#endif
	      /*
	       * Error 
	       */
	      syslog (LOG_USER | LOG_WARNING,
		      "Problem - command had checksum error.");
	      errorcounter++;
	    }
	  else if (recbuffer[2] == 0x16)
	    {
#ifdef DEBUG
	      printf ("NAK2 received\n");
#endif
	      /*
	       * Error 
	       */
	      syslog (LOG_USER | LOG_WARNING, "Problem - invalid command.\n");
	      errorcounter++;
	    }
	  else if ((recbuffer[2] >= 0x20) && (recbuffer[2] <= 0xff))
	    {
#ifdef DEBUG
	      printf ("Response value %x (%i) received.\n", recbuffer[2],
		      recbuffer[2]);
#endif
	      /*
	       * Check command, print value 
	       */
	      if (sendbuffer[3] == WATCHDOG_REV_MAJOR)
		{
		  recvalue = recbuffer[2] - 0x20;
		  syslog (LOG_USER | LOG_NOTICE,
			  "Firmware major rev. %i", recvalue);
		}
	      else if (sendbuffer[3] == WATCHDOG_REV_MINOR)
		{
		  recvalue = recbuffer[2] - 0x20;
		  syslog (LOG_USER | LOG_NOTICE,
			  "Firmware minor rev. %i", recvalue);
		}
	      else if (sendbuffer[3] == WATCHDOG_RESETCOUNT)
		{
		  recvalue = recbuffer[2] - 0x20;
		  syslog (LOG_USER | LOG_NOTICE, "Resetcount %i", recvalue);
		}
	      else if (sendbuffer[3] == WATCHDOG_TEMPERATURE)
		{
		  recvalue = recbuffer[2] - 0x40;
		  syslog (LOG_USER | LOG_NOTICE,
			  "Temperature %i C", (char) recvalue);
		}
	      else if (sendbuffer[3] == WATCHDOG_ARM)
		{
		  if (recvalue == 0x20)
		    {
		      syslog (LOG_USER | LOG_NOTICE, "Watchdog armed!");
		    }
		  else
		    {
		      syslog (LOG_USER | LOG_NOTICE,
			      "Watchdog was already armed.");
		    }
		}
	    }
	  else
	    {
#ifdef DEBUG
	      printf ("Unknown response received.\n");
#endif
	      syslog (LOG_USER | LOG_INFO, "Problem - unknown return code.");
	      errorcounter++;
	    }

	}			/*
				 * Readerror 
				 */

      /*
       * Check errorcounter - maybe break out of loop 
       */
      if (errorcounter > ERROR_LIMIT)
	{
#ifdef DEBUG
	  printf ("Errorlimit reached.\n");
#endif
	  syslog (LOG_USER | LOG_ERR, "Exiting - error limit reached.");
	  in_loop = 0;
	}

      /*
       * Update loopcounter and wait 
       */
      if (loopcounter == (LOOPS_BEFORE_STATUS + 1))
	{
	  /* Restart with temperature as status message */
	  loopcounter = 4;
	}
      else
	{
	  loopcounter++;
	}

      /*
       * Pause 
       */
      sleep (LOOP_DELAY);

    }

}


/*
 * Open and setup device 
 */

int
StartDevice (char *device, int *fd)
{
  struct termios newtio;

  /*
   * Open device 
   */
  *fd = open (device, O_RDWR | O_NOCTTY | O_SYNC);
  if (*fd < 0)
    {
      syslog (LOG_USER | LOG_ERR, "Exiting - cannot open device %s.", device);
      return (-1);
    }
#ifdef DEBUG
  printf ("Opened device %s, fd=%i\n", device, *fd);
#endif

  /*
   * Get current terminal setup 
   */
  tcgetattr (*fd, &oldtio);

  /*
   * Termial setup - 1200 8N2 
   */
  bzero (&newtio, sizeof (newtio));
  newtio.c_cflag = B1200 | CS8 | CLOCAL | CREAD | CSTOPB;
  newtio.c_iflag = IGNPAR;
  newtio.c_oflag = 0;
  newtio.c_lflag = 0;
  newtio.c_cc[VTIME] = 0;
  newtio.c_cc[VMIN] = 1;
  tcflush (*fd, TCIFLUSH);
  tcsetattr (*fd, TCSANOW, &newtio);

#ifdef DEBUG
  printf ("Set terminal to 1200 8N2\n");
#endif

  /*
   * OK 
   */
  return (0);
}

/*
 * Reset device 
 */

int
StopDevice (int fd)
{
  /*
   * Restore the old port settings 
   */
  tcsetattr (fd, TCSANOW, &oldtio);

  /*
   * Close device 
   */
  close (fd);

  syslog (LOG_USER | LOG_INFO, "Watchdog service stopped.", ERROR_LIMIT);

#ifdef DEBUG
  printf ("Closed device.\n");
#endif

  return (0);
}

/*
 * ============ Main program 
 */


int
main (int argc, char *argv[])
{
  int i;
  int fd;
  char devicename[MAX_FILENAME_LEN];


  /*
   * Setup default device 
   */
  strcpy (devicename, "/dev/ttyS0");

  /*
   * Parse command line parameters for local program 
   */
  for (i = 1; i < argc; i++)
    {
      char *arg;
      if (argv[i][0] != '-')
	break;
      arg = argv[i];
      arg++;
      if (*arg == '-')
	arg++;
      if (strncmp (arg, "device", 6) == 0)
	{
	  i++;
	  if (i <= argc)
	    {
	      strncpy (devicename, argv[i], MAX_FILENAME_LEN);
	    }
	}
      else if (strncmp (arg, "delay", 5) == 0)
	{
	  i++;
	  if (i <= argc)
	    {
	      LOOP_DELAY = abs (atoi (argv[i]));
	    }
	}
      else if (strncmp (arg, "status", 6) == 0)
	{
	  i++;
	  if (i <= argc)
	    {
	      LOOPS_BEFORE_STATUS = abs (atoi (argv[i]));
	    }
	}
      else if (strncmp (arg, "error", 5) == 0)
	{
	  i++;
	  if (i <= argc)
	    {
	      ERROR_LIMIT = abs (atoi (argv[i]));
	    }
	}
      else if (strncmp (arg, "reset", 5) == 0)
	{
	  RESET_NOW = 1;
	}
      else if (strncmp (arg, "help", 4) == 0)
	{
	  /*
	   * Print local help 
	   */
	  printf
	    ("dpwatch - Berkshire products watchdog driver\nA.Schiffler/Deskplayer Inc., aschiffler@home.com, 05/00\n\n");
	  printf ("\nDevice settings:\n");
	  printf (" -device dev    watchdog device (default: %s)\n",
		  devicename);
	  printf ("\nRuntime settings:\n");
	  printf
	    (" -delay #       # seconds delay between commands (default: %i)\n",
	     LOOP_DELAY);
	  printf
	    (" -status #      # loops between status messages (default: %i)\n",
	     LOOPS_BEFORE_STATUS);
	  printf
	    (" -error #       tolerate # errors before exiting (default: %i)\n",
	     ERROR_LIMIT);
	  printf ("\nSpecial commands:\n");
	  printf
	    (" -reset         reset computer after 10 seconds (use with caution)\n");
	  /*
	   * Leave 
	   */
	  exit (0);
	}
    }

  /*
   * Open syslog 
   */
  openlog ("dpwatch", LOG_NDELAY, LOG_USER);

  /*
   * Message 
   */
  syslog (LOG_USER | LOG_INFO,
	  "dpwatch Ver. 1.0 - A. Schiffler/Deskplayer Inc., 05/00");
  syslog (LOG_USER | LOG_INFO,
	  "Berkshire Products Internal Serial Watchdog service started.");
  syslog (LOG_USER | LOG_INFO, "Serial Device: %s", devicename);
  syslog (LOG_USER | LOG_INFO, "Loop Delay:    %i sec", LOOP_DELAY);
  syslog (LOG_USER | LOG_INFO, "Status after:  %i loops", LOOPS_BEFORE_STATUS);
  syslog (LOG_USER | LOG_INFO, "Error-Limit:   %i errors", ERROR_LIMIT);

  /* Setup signal for program shutdown through Ctrl-C */
  signal (SIGINT, &Quit);

  /* Setup signal for program shutdown through 'killall' */
  signal (SIGTERM, &Quit);

  /*
   * Do it 
   */
  if (StartDevice (devicename, &fd) < 0)
    exit (-1);
  ResetLoop (fd);
  if (StopDevice (fd) < 0)
    exit (-1);

  exit (0);
}
